# -*- coding: UTF-8 -*-
# python-wifi -- a wireless library to access wireless cards via python
# Version: 0.3.1
# Copyright (C) 2004 - 2008 Róman Joost
#
# Contributions from:
#   Mike Auty <m.auty@softhome.net> (Iwscanresult, Iwscan)
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU General Public License
#    as published by the Free Software Foundation; either version 2.0 of
#    the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful, but
#    WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#    General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this library; if not, write to the Free Software Foundation, Inc.,
#    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
#
#   This library was modified by Fernando Pablo Romero Navarro (May-2010):
#    * Some bugs have been fixed.
#    * Some functions have been modified.
#    * Added new "Wireless Extensions" functions: 
#                                       SetFrequency, SetChannel, SetBitrate
#    * Added new functions for enabling/disabling the interface and checking
#      whether it is enabled.
#    * Added new functions for getting the MAC address of the interface and
#      setting the IP address.

import struct
import array
import math
import fcntl
import socket
import time
import re
import flags



KILO = 10**3
MEGA = 10**6
GIGA = 10**9


def getNICnames():  
    device = re.compile('[a-z]{2,4}[0-9]') 
    ifnames = []
    
    fp = open('/proc/net/wireless', 'r')
    data = fp.readlines()
    for line in data:
        try:
            ifnames.append(device.search(line).group())
        except AttributeError:
            pass 
    
    if ifnames == []:
        ifnames = getConfiguredNICnames()
    
    return ifnames


def getConfiguredNICnames():
    
    iwstruct = Iwstruct()
    ifnames = []
    buff = array.array('c', '\0'*1024)
    caddr_t, length = buff.buffer_info()
    datastr = iwstruct.pack('iP', length, caddr_t)
    try:
        result = iwstruct._fcntl(flags.SIOCGIFCONF, datastr)
    except IOError, (i, error):
        return i, error

    
    for i in range(0, 1024, 32):
        ifname = buff.tostring()[i:i+32]
        ifname = struct.unpack('32s', ifname)[0]
        ifname = ifname.split('\0', 1)[0]
        if ifname:
            
            wifi = Wireless(ifname)
            result = wifi.getAPaddr()
            if result[0] == 0:
                ifnames.append(ifname)

    return ifnames  


def makedict(**kwargs):
    return kwargs


class Wireless(object):
    
    def __init__(self, ifname):        
        self.ifname = ifname
        self.iwstruct = Iwstruct()

    def getAPaddr(self):        
        _ , datastr = self.iwstruct.pack_wrq(32)
        status, result = self.iwstruct.iw_get_ext(self.ifname, 
                                                  flags.SIOCGIWAP,
                                                  data=datastr)
        if status > 0:
            return (status, result)

        return self.iwstruct.getMAC(result)

    def getHWAddress(self):        
           
        st,res=self.iwstruct.iw_get_ext(self.ifname, flags.SIOCGIFHWADDR)
        if(st!=0):
            raise EnvironmentError(st, res)
        else:
            fmt="@h6s"
            lng=len(res)-struct.calcsize(fmt)
            fmt=fmt+("%dx" % lng)
            
        return struct.unpack(fmt, res)

    def is_enabled(self):
        ifflags=0

        st,res=self.iwstruct.iw_get_ext(self.ifname, flags.SIOCGIFFLAGS)

        if(st!=0):
            raise EnvironmentError(st, res)
        else:
            fmt="@h"
            lng=len(res)-struct.calcsize(fmt)
            fmt=fmt+("%dx" % lng)
            ifflags=struct.unpack(fmt, res)[0]

        return (ifflags & flags.IFF_UP!=0)


    def set_enable(self, enable):       
        change=False
        ifflags=0
        st,res=self.iwstruct.iw_get_ext(self.ifname, flags.SIOCGIFFLAGS)

        if(st!=0):
            raise EnvironmentError(st, res)
        else:
            fmt="@h"
            lng=len(res)-struct.calcsize(fmt)
            fmt=fmt+("%dx" % lng)
            ifflags=struct.unpack(fmt, res)[0]

        enabled=(ifflags & flags.IFF_UP)!=0

        if enabled and not enable:
            change=True
            ifflags=ifflags & ~flags.IFF_UP
        elif not enabled and enable:
            change=True
            ifflags=ifflags | flags.IFF_UP

        if change:
            ifreq=struct.pack("@h", ifflags)
            st, res=self.iwstruct.iw_set_ext(self.ifname,
                                             flags.SIOCSIFFLAGS,
                                             ifreq)
        else:
            st=0

        if st!=0:
            raise EnvironmentError(st, res)
        
    def set_address(self, ipaddr):
        rawip=socket.inet_aton(ipaddr)
        ifreq=struct.pack("HH4s", socket.AF_INET, 0, rawip)
        st, res=self.iwstruct.iw_set_ext(self.ifname, flags.SIOCSIFADDR, ifreq)
        
        if st!=0:
            raise EnvironmentError(st, res)


    def release(self):       
        self.ifname=None
        self.iwstruct.sockfd.close()
        self.iwstruct=None

   
    def getBitrate(self):
        
        i, result = self.iwstruct.iw_get_ext(self.ifname, 
                                            flags.SIOCGIWRATE)
        if i > 0:
            return (i, result)
        iwfreq = Iwfreq(result)
        return iwfreq.getBitrate()
    
    def getBitrates(self):
        
        iwrange = Iwrange(self.ifname)
        if iwrange.errorflag:
            return (iwrange.errorflag, iwrange.error)
        return (iwrange.num_bitrates, iwrange.bitrates)
    
    def setBitrate(self, mbits):
        wreq=struct.pack("iBBH", mbits*10**6, 1, 0, 0)
        i, result=self.iwstruct.iw_set_ext(self.ifname, flags.SIOCSIWRATE, wreq)
        
        if i>0:
            return (i, result)           

    def getChannelInfo(self):
        
        iwrange = Iwrange(self.ifname)
        if iwrange.errorflag:
            return (iwrange.errorflag, iwrange.error)
        return (iwrange.num_channels, iwrange.frequencies)

    def getEssid(self):        
        essid = ""
        buff, s = self.iwstruct.pack_wrq(32)
        i, result = self.iwstruct.iw_get_ext(self.ifname, 
                                             flags.SIOCGIWESSID, 
                                             data=s)
        if i > 0:
            return (i, result)
        datastr = buff.tostring()
        return datastr.strip('\x00')

    def setEssid(self, essid):
        
        if len(essid) > flags.IW_ESSID_MAX_SIZE:
            return "essid to big"
        _, datastr = self.iwstruct.pack_test(essid, 32)
        status, result = self.iwstruct.iw_set_ext(self.ifname,
                                             flags.SIOCSIWESSID, 
                                             data=datastr)
        if status > 0:
            return (status, result)

    def getEncryption(self):

        iwpoint = Iwpoint(self.ifname)
        if iwpoint.errorflag:
            return (iwpoint.errorflag, iwpoint.error)
        return iwpoint.getEncryptionKey()

    def getFragmentation(self):
        
        iwparam = Iwparam(self.ifname, flags.SIOCGIWFRAG)
        if iwparam.errorflag:
            return (iwparam.errorflag, iwparam.error)
        return iwparam.getValue()

    def getFrequency(self):
        
        status, result = self.iwstruct.iw_get_ext(self.ifname, 
                                                  flags.SIOCGIWFREQ)
        if status > 0:
            return (status, result)
        iwfreq = Iwfreq(result)
        return iwfreq.tostring()

    def setFrequency(self, freq):        
        msjErr="Invalid Argument"
        val=None
        mant=None
        exp=None
        
        if(type(freq) is int or float):
            val=float(freq)
        elif(type(freq) is str):
            try:
                lng=len(freq)
                suffix=freq[lng-1]
                if suffix=='K':
                    mult=10**3
                elif suffix=='M':
                    mult=10**6
                elif suffix=='G':
                    mult=10**9
                else:
                    raise ValueError, msjErr

                val=float(freq[:lng-1])*mult
            except ValueError:
                raise ValueError, msjErr
        else:
            raise ValueError, msjErr

        if(val<2.0e+09 or val>8.0e+09):
            raise ValueError, "Argument out of range"
        else:
            exp=6
            mant=math.floor(val/10**6)
            wrq=struct.pack("@ihBB", mant, exp, 0, flags.IW_FREQ_FIXED)

        status, result=self.iwstruct.iw_set_ext(self.ifname,
                                                flags.SIOCSIWFREQ,
                                                data=wrq)

        return (status, result)


    def setChannel(self, chan):
        iwrange = Iwrange(self.ifname)
        
        if(chan<=0  or chan>iwrange.num_frequency):
            raise ValueError, "Argument out of range"
        else:
            freq=iwrange.frequencies[chan-1]
            self.setFrequency(freq)
            

    def getMode(self):        
        status, result = self.iwstruct.iw_get_ext(self.ifname, 
                                             flags.SIOCGIWMODE)
        if status > 0:
            return (status, result)
        mode = self.iwstruct.unpack('B', result[0])[0]
        return flags.modes[mode]

    def setMode(self, mode):
        
        try:
            this_modes = [x.lower() for x in flags.modes]
            mode = mode.lower()
            wifimode = this_modes.index(mode)
        except ValueError:
            return "Invalid operation mode!"

        datastr = self.iwstruct.pack('I', wifimode)
        status, result = self.iwstruct.iw_set_ext(self.ifname,
                                                  flags.SIOCSIWMODE,
                                                  data=datastr)
        if status > 0:
            return (status, result)

    def getWirelessName(self):
        
        status, result = self.iwstruct.iw_get_ext(self.ifname, 
                                             flags.SIOCGIWNAME)
        if status > 0:
            return (status, result)
        return result.split('\0')[0]

    def getPowermanagement(self):
       
        iwparam = Iwparam(self.ifname, flags.SIOCGIWPOWER)
        if iwparam.errorflag:
            return (iwparam.errorflag, iwparam.error)
        return iwparam.getValue()

    def getRetrylimit(self):
        
        iwparam = Iwparam(self.ifname, flags.SIOCGIWRETRY)
        if iwparam.errorflag:
            return (iwparam.errorflag, iwparam.error)
        return iwparam.getValue()

    def getRTS(self):
        
        iwparam = Iwparam(self.ifname, flags.SIOCGIWRTS)
        if iwparam.errorflag:
            return (iwparam.errorflag, iwparam.error)
        return iwparam.getValue()

    def getSensitivity(self):
        
        iwparam = Iwparam(self.ifname, flags.SIOCGIWSENS)
        if iwparam.errorflag:
            return (iwparam.errorflag, iwparam.error)
        return iwparam.getValue()

    def getTXPower(self):
        
        status, result = self.iwstruct.iw_get_ext(self.ifname, 
                                                  flags.SIOCGIWTXPOW)
        if status > 0:
            return (status, result)
        iwfreq = Iwfreq(result)
        return iwfreq.getTransmitPower()

    def getStatistics(self):
        
        iwstats = Iwstats(self.ifname)
        if iwstats.errorflag > 0:
            return (iwstats.errorflag, iwstats.error)
        return [iwstats.status, iwstats.qual, iwstats.discard,
            iwstats.missed_beacon]

    def scan(self):        
        iwscan = Iwscan(self.ifname)
        return iwscan.scan()


class Iwstruct(object):
    
    def __init__(self):
        self.idx = 0
        self.sockfd = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def parse_data(self, fmt, data):

        size = struct.calcsize(fmt)
        idx = self.idx

        datastr = data[idx:idx + size]
        self.idx = idx+size
        value = struct.unpack(fmt, datastr)


        if len(value) == 1:
            return value[0]
        else:
            return value
    
    def pack(self, fmt, *args):        
        return struct.pack(fmt, *args)

    def pack_wrq(self, buffsize):        
        buff = array.array('c', '\0'*buffsize)
        caddr_t, length = buff.buffer_info()
        datastr = struct.pack('Pi', caddr_t, length)
        return buff, datastr
    
    def pack_test(self, string, buffsize):        
        buffsize = buffsize - len(string)
        buff = array.array('c', string+'\0'*buffsize)
        caddr_t, length = buff.buffer_info()
        s = struct.pack('Pii', caddr_t, length, 1)
        return buff, s

    def unpack(self, fmt, packed_data):        
        return struct.unpack(fmt, packed_data)

    def _fcntl(self, request, args):
        return fcntl.ioctl(self.sockfd.fileno(), request, args)
    
    def iw_get_ext(self, ifname, request, data=None):
        
        if data is not None:
            buff = flags.IFNAMSIZE-len(ifname)
            ifreq = ifname + '\0'*buff
            ifreq = ifreq + data
        else:
            ifreq = (ifname + '\0'*32)

        try:
            result = self._fcntl(request, ifreq)
        except IOError, (i, e):
            return i, e

        return (0, result[16:])

    def iw_set_ext(self, ifname, request, data):

        if data is not None:
            buff = flags.IFNAMSIZE-len(ifname)
            ifreq = ifname + '\0'*buff
            ifreq = ifreq + data
            try:
                result = self._fcntl(request, ifreq)
            except IOError, (i, e):
                return i, e

            return (0, result[16:])


    def getMAC(self, packed_data):
        mac_addr = struct.unpack('xxBBBBBB', packed_data[:8])
        return "%02x:%02x:%02x:%02x:%02x:%02x" % mac_addr


class Iwparam(object):
    
    def __init__(self, ifname, ioctl):        
        self.fmt = "ibbH"
        self.value = 0
        self.fixed = 0
        self.disabled = 0
        self.flags = 0
        self.errorflag = 0
        self.error = ""
        self.ioctl = ioctl 
        self.ifname = ifname
        self.update()
    
    def getValue(self):        

        if self.disabled:
            return 'off'
        if self.flags & flags.IW_RETRY_TYPE == 0:
            return self.getRLAttributes()
        else:
            return self.getPMAttributes()

    def getRLAttributes(self):
        return self.value

    def getPMAttributes(self):
        result = ""        
        
        if self.flags & flags.IW_POWER_MIN == 0:
            result = " min"
        if self.flags & flags.IW_POWER_MAX == 0:
            result = " max"            
        
        if self.flags & flags.IW_POWER_TIMEOUT == 0:
            result = " period:" 
        else:
            result = " timeout:"
        
        if self.flags & flags.IW_POWER_RELATIVE:
            result += "%f" % (float(self.value)/MEGA)
        else:
            if self.value >= MEGA:
                result += "%fs" % (float(self.value)/MEGA)
            elif self.value >= KILO:
                result += "%fms" % (float(self.value)/KILO)
            else:
                result += "%dus" % self.value

        return result
        
    def update(self):
        
        iwstruct = Iwstruct()
        i, r = iwstruct.iw_get_ext(self.ifname, 
                                   self.ioctl)
        if i > 0:
            self.errorflag = i
            self.error = r
        self._parse(r)
    
    def _parse(self, data):
        
        iwstruct = Iwstruct()
        self.value, self.fixed, self.disabled, self.flags =\
            iwstruct.parse_data(self.fmt, data)
        
class Iwfreq(object):    
    
    def __init__(self, data=None, _range=None):
        
        self.fmt = "ihbb"
        self.frequency=0
        self.channel=0
        self.range=_range
        
        if data is not None:
            self.parse(data)

        self.iwstruct = Iwstruct()
        
    def __getattr__(self, attr):
        return getattr(self.iwstruct, attr)

    def parse(self, data):
        
        size = struct.calcsize(self.fmt)
        m, e, i, _ = struct.unpack(self.fmt, data[:size])
        
        if m<=Iwrange.IW_MAX_FREQUENCIES:
            self.channel=m
            if(self.range is not None):
                numfreqs=len(self.range.frequencies)
                ind=((m < numfreqs) and m) or (numfreqs-1)
                self.frequency=self.range.frequencies[ind]
        else:            
            
            if e == 0:
                self.frequency=m
            else:
                self.frequency=float(m)*10**e

            if(self.range is not None):
                ind=0
                dif="+inf"
                tmp=0
                i=0
                for freq in self.range.frequencies:
                    tmp=abs(self.frequency-freq)
                    i=i+1
                    if(dif > tmp):
                        dif=tmp
                        ind=i

                self.channel=ind

    def getFrequency(self):        
        return self.frequency

    
    def tostring(self):        
        freq = self.frequency
        
        if freq >= GIGA:
            return "%0.3fGHz" % (freq/GIGA)

        if freq >= MEGA:
            return "%0.3fMHZ" % (freq/MEGA)

        if freq >= KILO:
            return "%0.3fKHz" % (freq/KILO)

    def getBitrate(self):        
        bitrate = self.frequency

        if bitrate >= GIGA:
            return "%i Gb/s" % (bitrate/GIGA)

        if bitrate >= MEGA:
            return "%i Mb/s" % (bitrate/MEGA)

        if bitrate >= KILO:
            return "%i Kb/s" % (bitrate/KILO)

    def getTransmitPower(self):        
        return "%i dBm" %self.mw2dbm(self.frequency/10)

    def getChannel(self):        
        return self.channel


    def mw2dbm(self, mwatt):        
        return math.ceil(10.0 * math.log10(mwatt))

    def _setFrequency(self, vallist):        
        assert len(vallist) == 4
        m, e, i, pad = vallist
        if e == 0:
            self.frequency = m
        else:
            self.frequency = float(m)*10**e


class Iwstats(object):    

    def __init__(self, ifname):
        self.fmt = "2B4B6i"
        self.status = 0
        self.qual = Iwquality()
        self.discard = {}
        self.missed_beacon = 0
        self.ifname = ifname
        self.errorflag = 0
        self.error = ""
        self.update()

    def update(self):        
        iwstruct = Iwstruct()
        buff, s = iwstruct.pack_wrq(32)
        i, result = iwstruct.iw_get_ext(self.ifname, 
                                        flags.SIOCGIWSTATS, 
                                        data=s)
        if i > 0:
            self.error = result
            self.errorflag = i
        self._parse(buff.tostring())
    
    def _parse(self, data):        
        iwstruct = Iwstruct()
        iwqual = Iwquality()
        iwstats_data = iwstruct.parse_data(self.fmt, data)
        
        self.status = iwstats_data[0:2]
        self.qual.quality, self.qual.siglevel, self.qual.nlevel, \
            self.qual.updated = iwstats_data[2:6]
        nwid, code, frag, retries, iwflags = iwstats_data[6:11]
        self.missed_beacon = iwstats_data[11:12][0]
        self.discard = makedict(nwid=nwid, code=code,
            fragment=frag, retries=retries, misc=iwflags)


class Iwquality(object):

    def __init__(self):
        self.quality = 0
        self.siglevel = 0
        self.nlevel = 0
        self.updated = 0
        self.fmt = "4B"

    def parse(self, data):
        iwstruct = Iwstruct()
        qual, siglevel, nlevel, iwflags = iwstruct.parse_data(self.fmt, data)

        self.siglevel = siglevel
        self.nlevel = nlevel

        self.quality = qual
        self.updated = iwflags

    def setValues(self, vallist):        
        attributes = ["quality", "signallevel", "noise_level",
            "updated"]
        assert len(vallist) == 4
        
        for i in range(len(vallist)):
            setattr(self, attributes[i], vallist[i])
    
    def getSignallevel(self):       
        return self.siglevel-0x100

    def setSignallevel(self, siglevel):        
        self.siglevel = siglevel
        
    signallevel = property(getSignallevel, setSignallevel)
    
    def getNoiselevel(self):        
        return self.nlevel - 0x100

    def setNoiselevel(self, val):        
        self.nlevel = val

    noiselevel = property(getNoiselevel, setNoiselevel)


class Iwpoint(object):    

    def __init__(self, ifname):
        self.key = [0, 0, 0, 0]
        self.fields = 0
        self.flags = 0        
        self.fmt = "4BHH"
        self.errorflag = 0
        self.error = ""
        self.ifname = ifname
        self.update()

    def __getattr__(self, attr):
        return getattr(self.iwstruct, attr)
    
    def update(self):        
        iwstruct = Iwstruct()
        _ , datastr = iwstruct.pack_wrq(32)
        i, result = iwstruct.iw_get_ext(self.ifname, 
                                        flags.SIOCGIWENCODE, 
                                        data=datastr)
        if i > 0:
            self.errorflag = i
            self.error = result
            
        self._parse(result)
        
    def getEncryptionKey(self):        
        if self.flags & flags.IW_ENCODE_DISABLED != 0:
            return 'off'
        elif self.flags & flags.IW_ENCODE_NOKEY != 0:            
            return '**' * self.fields
    
    def _parse(self, data):
        
        iwstruct = Iwstruct()
        ptr, ptr, ptr, ptr, self.fields, self.flags = \
            iwstruct.parse_data(self.fmt, data)
        self.key = [ptr, ptr, ptr, ptr]


class Iwrange(object):    
    IW_MAX_FREQUENCIES = 32

    def __init__(self, ifname):
        self.fmt = "iiihb6ii4B4Bi32i2i2i2i2i3h8h2b2bhi8i2b3h2i2ihBx"\
            + self.IW_MAX_FREQUENCIES*"ihbb"
        
        self.ifname = ifname
        self.errorflag = 0
        self.error = ""
        
        
        self.throughput = 0        
        self.min_nwid = self.max_nwid = 0                
        self.old_num_channels = self.old_num_frequency = self.old_freq = 0                
        self.sensitivity = 0
        self.max_qual = Iwquality()
        self.avg_qual = Iwquality()        
        self.num_bitrates = 0
        self.bitrates = []
        
        self.min_rts = self.max_rts = 0       
        self.min_frag = self.max_frag = 0        
        self.min_pmp = self.max_pmp = 0
        self.min_pmt = self.max_pmt = 0
        self.pmp_flags = self.pmt_flags = self.pm_capa = 0
        
        self.encoding_size = 0
        self.num_encoding_sizes = self.max_encoding_tokens = 0
        self.encoding_login_index = 0        
        self.txpower_capa = self.num_txpower = self.txpower = 0        
        self.we_vers_compiled = self.we_vers_src = 0
        
        self.retry_capa = self.retry_flags = self.r_time_flags = 0
        self.min_retry = self.max_retry = 0
        self.min_r_time = self.max_r_time = 0

        self.num_channels = self.num_frequency = 0
        self.frequencies = []
        self.update()
    
    def update(self):
        
        iwstruct = Iwstruct()
        buff, s = iwstruct.pack_wrq(640)
        i, result = iwstruct.iw_get_ext(self.ifname, 
                                        flags.SIOCGIWRANGE, 
                                        data=s)
        if i > 0:
            self.errorflag = i
            self.error = result
        data = buff.tostring()
        self._parse(data)
        
    def _parse(self, data):
        iwstruct = Iwstruct()
        result = iwstruct.parse_data(self.fmt, data)        
        
        self.throughput, self.min_nwid, self.max_nwid = result[0:3]
        self.old_num_channels, self.old_num_frequency = result[3:5]
        self.old_freq = result[5:11]
        self.sensitivity = result[11]
        self.max_qual.setValues(result[12:16])
        self.avg_qual.setValues(result[16:20])
        self.num_bitrates = result[20]
        raw_bitrates = result[21:53]

        for rate in raw_bitrates:
            iwfreq = Iwfreq()
            iwfreq.frequency = rate
            btr = iwfreq.getBitrate()
            if btr is not None:
                self.bitrates.append(btr)
            
        self.min_rts, self.max_rts = result[53:55]
        self.min_frag, self.max_frag = result[55:57]
        self.min_pmp, self.max_pmp = result[57:59]
        self.min_pmt, self.max_pmt = result[59:61]
        self.pmp_flags, self.pmt_flags, self.pm_capa = result[61:64]
        self.encoding_size = result[64:72]
        self.num_encoding_sizes, self.max_encoding_tokens = result[72:74]
        self.encoding_login_index = result[74:76]
        self.txpower_capa, self.num_txpower = result[76:78]
        self.txpower = result[78:86]
        self.we_vers_compiled, self.we_vers_src = result[86:88]
        self.retry_capa, self.retry_flags, self.r_time_flags = result[88:91]
        self.min_retry, self.max_retry = result[91:93]
        self.min_r_time, self.max_r_time = result[93:95]
        self.num_channels = result[95]
        self.num_frequency = result[96]
        freq = result[97:]
        
        i = self.num_frequency
        for x in range(0, len(freq), 4):
            iwfreq = Iwfreq()
            iwfreq._setFrequency(freq[x:x+4])
            fq = iwfreq.getFrequency()
            if fq is not None:
                self.frequencies.append(fq)
            i -= 1
            if i <= 0:
                break
                
        self.frequencies.sort()

class Iwscan(object):
    
    def __init__(self, ifname):
        self.ifname = ifname
        self.range = Iwrange(ifname)
        self.errorflag = 0
        self.error = ""
        self.stream = None
        self.aplist = None
                
    def scan(self, fullscan=True):        
        result = 1
        if fullscan:
            self.setScan()
            if self.errorflag > flags.EPERM:
                errormsg = "setScan failure %s %s" % (str(self.errorflag),
                                                   str(self.error))
                raise RuntimeError(errormsg)
                return None
            elif self.errorflag < flags.EPERM:            
                result = 250
        
        while (result > 0):
            time.sleep(result/1000)
            result = self.getScan()
        
        if result < 0 or self.errorflag != 0:
            raise RuntimeError, 'getScan failure ' + str(self.errorflag)\
                                + " " + str(self.error)
        
        return self.aplist
        
        
    def setScan(self):        
        iwstruct = Iwstruct()
        datastr = iwstruct.pack('Pii', 0, 0, 0)
        status, result = iwstruct.iw_get_ext(self.ifname, 
                                             flags.SIOCSIWSCAN, datastr)
        if status > 0:
            self.errorflag = status
            self.error = result
        return result
        
    def getScan(self):        
        iwstruct = Iwstruct()
        i = flags.E2BIG
        bufflen = flags.IW_SCAN_MAX_DATA
        
        while (i == flags.E2BIG):
            buff, datastr = iwstruct.pack_wrq(bufflen)
            i, result = iwstruct.iw_get_ext(self.ifname, 
                                            flags.SIOCGIWSCAN,
                                            data=datastr)
            if i == flags.E2BIG:
                _, newlen = iwstruct.unpack('Pi', datastr)
                if bufflen < newlen:
                    bufflen = newlen
                else:
                    bufflen = bufflen * 2

        if i == flags.EAGAIN:
            return 100
        if i > 0:
            self.errorflag = i
            self.error = result
            return -1

        _, reslen = iwstruct.unpack('Pi', datastr)
        if reslen > 0:
            self.aplist = self._parse(buff.tostring())
            return 0

    def _parse(self, data):        
        iwstruct = Iwstruct()
        scanresult = None
        aplist = []
        
        while 1:            
            if len(data) < flags.IW_EV_LCP_LEN:
                break;
            
            length, cmd = iwstruct.unpack('HH', data[:4])            
            if length < flags.IW_EV_LCP_LEN:
                break;
            
            if cmd == flags.SIOCGIWAP:
                if scanresult is not None:
                    aplist.append(scanresult)
                scanresult = Iwscanresult(
                    data[flags.IW_EV_LCP_LEN:length], self.range)
            elif scanresult is None:
                raise RuntimeError, 'Attempting to add an event without AP data'
            else:
                scanresult.addEvent(cmd,
                                    data[flags.IW_EV_LCP_LEN:length])
            
            data = data[length:]

        if scanresult is None:
            return None         
        
        if scanresult.bssid != "00:00:00:00:00:00":
            aplist.append(scanresult)
        else:
            raise RuntimeError, 'Attempting to add an AP without a bssid'
        return aplist


class Iwscanresult(object): 
    
    def __init__(self, data, iwrange):
        self.iwstruct = Iwstruct()
        self.range = iwrange
        self.bssid = "%02x:%02x:%02x:%02x:%02x:%02x" % (
            struct.unpack('BBBBBB', data[2:8]))
        self.essid = None
        self.mode = None
        self.rate = []
        self.quality = Iwquality() 
        self.frequency = None
        self.encode = None
        self.custom = []
        self.protocol = None
        

    def addEvent(self, cmd, data):        
        if cmd <= flags.SIOCIWLAST:
            if cmd < flags.SIOCIWFIRST:
                return None
        elif cmd >= flags.IWEVFIRST:
            if cmd > flags.IWEVLAST:
                return None
        else:
            return None
            
        if cmd == flags.SIOCGIWESSID:
            self.essid = data[4:]
        elif cmd == flags.SIOCGIWMODE:
            self.mode = flags.modes[self.iwstruct.unpack('i', data[:4])[0]]
        elif cmd == flags.SIOCGIWRATE:            
            freqsize = struct.calcsize("ihbb")
            while len(data) >= freqsize:
                iwfreq = Iwfreq(data)
                self.rate.append(iwfreq.getBitrate())
                data = data[freqsize:]
        elif cmd == flags.IWEVQUAL:
            self.quality.parse(data)
        elif cmd == flags.SIOCGIWFREQ:
            self.frequency = Iwfreq(data, self.range)
        elif cmd == flags.SIOCGIWENCODE:
            lastbyte=ord(data[len(data)-1])
            if lastbyte==0x80:
                self.encode = flags.IW_ENCODE_NONE
            elif lastbyte==0x08:
                if self.encode is None:
                    self.encode=flags.IW_ENCODE_WEP
        elif cmd == flags.IWEVCUSTOM:
            self.custom.append(data[1:])
        elif cmd == flags.SIOCGIWNAME:
            self.protocol = data[:len(data)-2]
        elif cmd==flags.IWEVGENIE:
            offset=flags.IW_EV_POINT_LEN - flags.IW_EV_LCP_LEN
            if ord(data[offset])==flags.IW_RSN_IE_ID:
                self.encode=flags.IW_ENCODE_WPA
            elif ord(data[offset])==flags.IW_MULT_IE_ID:
                if(data[offset+2:offset+6]==flags.IW_WPA_OUI_IE):
                    self.encode=flags.IW_ENCODE_WPA

        else:
            print "Cmd:", cmd
            return False
        return True

    def display(self):
        print "ESSID: ", self.essid
        print "Access point: ", self.bssid
        print "Mode: ", self.mode
        if len(self.rate) > 0:
            print "Highest Bitrate: ", self.rate[len(self.rate)-1]
        print "Quality: ", self.quality.quality,
        print "Signal: ", self.quality.getSignallevel(),
        print "Noise: ", self.quality.getNoiselevel()
        print "Frequency: ", self.frequency.tostring()
        print "Channel: ", self.frequency.getChannel()
       
        if self.encode is not None:
            if self.encode==flags.IW_ENCODE_NONE:
                enc_type="None"
            elif self.encode==flags.IW_ENCODE_WEP:
                enc_type="WEP"
            elif self.encode==flags.IW_ENCODE_WPA:
                enc_type="WPA"

            print "Encription: ", enc_type
        
        for custom in self.custom:
            print "Custom:", custom
        print ""
